@@ -22,7 +22,7 @@ class Object
   private
 
   def bug!( message = 'must not happen' )
-    raise FindBug, "\n\n[SCRIPT BUG] " + message
+    raise FindBug, "\n[SCRIPT BUG] " + message
   end
 
 end
