@@ -8,8 +8,9 @@
 #   the GNU Lesser General Public License version 2 or later.
 #
 
-require 'amstd/futils'
 require 'socket'
+require 'final'
+require 'amstd/futils'
 
 
 module TMail
@@ -26,18 +27,52 @@ module TMail
       @verbose = false
     end
 
+    attr :dirname
+
+    def each_mail
+      each_filename do |fname|
+        yield FilePort.new( fname )
+      end
+      @lasttime = Time.now
+    end
+
+    alias each each_mail
+
+    def new_mail
+      FilePort.new nextfile
+    end
+
+    alias new_port new_mail
+
+    def each_newmail( mtime = nil, &block )
+      mtime ||= @lasttime
+      unless mtime then
+        return each_mail( &block )
+      end
+
+      return if mtime > File.mtime( @dirname )
+      each_filename do |fname|
+        if File.mtime( fname ) > mtime then
+          yield FilePort.new( fname )
+        end
+      end
+      @lasttime = Time.now
+    end
+
+
+    private
 
     def each_filename( &block )
       ret = []
-      foreach_fullpath( @dirname ) do |path|
-        next unless /\A\d+\z/ === basename(path)
+      Dir.foreach( @dirname ) do |fn|
+        path = "#{@dirname}/#{fn}"
+        next unless /\A\d+\z/ === fn
         next unless file? path
         ret.push path
       end
       ret.each( &block )
     end
 
-
     def nextfile
       n = @oldnext
       unless n then
@@ -54,37 +89,7 @@ module TMail
       end
       @oldnext = n
 
-      return "#{@dirname}/#{n}"
-    end
-
-
-    def each_mail
-      arr = []
-      @lasttime = Time.now
-      each_filename do |fname|
-        yield FilePort.new( fname )
-      end
-    end
-
-    alias each each_mail
-
-    def new_mail
-      return FilePort.new( nextfile )
-    end
-
-    alias new_port new_mail
-
-    def each_newmail( mtime = nil )
-      mtime ||= @lasttime
-      raise ArgumentError, "last update time is not given"
-
-      return if mtime > File.mtime( @dirname )
-      @lasttime = Time.now
-      each_filename do |fname|
-        if File.mtime( fname ) > mtime then
-          yield FilePort.new( fname )
-        end
-      end
+      "#{@dirname}/#{n}"
     end
 
   end
@@ -93,100 +98,142 @@ module TMail
 
   class MboxLoader
 
-    def initialize( fname, tempdir = nil )
-      @filename = expand( fname )
+    include FileUtils
+
+    def initialize( fname, dir = nil )
+      @filename = File.expand_path( fname )
       unless file? @filename then
         raise ArgumentError, "'#{fname}' is not normal file"
       end
 
-      if tempdir then
-        @tempdir = mustdir( tempdir )
+      if dir then
+        tmp = mustdir( dir )
       else
-        @tempdir = isdir( "/tmp/tmail_mboxloader_#{$$}_#{id}" )
-        @finalizer = MboxLoader.mkproc( @tempdir )
-        ObjectSpace.add_finalizer @finalizer
+        tmp = isdir( "/tmp/tmail_mboxloader_#{$$}_#{id}" )
       end
 
-      @real = nil
       @closed = false
+      @real = MhLoader.new( tmp )
+
+      @finalizer = MboxLoader.mkfinal( @real, @filename )
+      ObjectSpace.define_finalizer self, @finalizer
+    end
+
+    class << self
+      def mkfinal( ld, mbfile )
+        proc {
+          File.open( mbfile, 'w' ) do |f|
+            ld.each do |port|
+              fr = get_fromline( port )
+              port.ropen do |s|
+                f.puts fr
+                f.puts s.read_all
+              end
+            end
+          end
+          FILEUTILS.rm_rf ld.dirname
+        }
+      end
+
+      def get_fromline( port )
+        ad = 'nobody'
+        if h = HeaderField.new_header( port, 'From' ) then
+          if a = h.addrs[0] then
+            ad = a.address
+          end
+        end
+
+        t = File.mtime( port.filename )
+        t = t.strftime( '%a %b %d %H:%M:%S %Y' )
+
+        "From #{ad} #{t}"
+      end
     end
 
     def close
       return if @closed
-
       @closed = true
-      @real = nil
-      @tempdir = nil
+      
+      ObjectSpace.undefine_finalizer self
       @finalizer.call
       @finalizer = nil
+
+      @real = nil
     end
 
-
-    def each_mail
+    def each_mail( &block )
       close_check
-      sync
-      @real.each_mail {|p| yield p }
+      update
+      @real.each_mail( &block )
     end
 
     alias each each_mail
 
     def each_newmail( mtime = nil )
       close_check
-      sync
+      update
       @real.each_newmail( mtime ) {|p| yield p }
     end
 
-    def new_mail
-      raise
+    def new_port
+      close_check
+      @real.new_port
     end
 
-    alias new_port new_mail
+    alias new_mail new_port
 
 
     private
 
-    def sync
-      wf = nil
-      n = 1
+    FROMLINE = /\AFrom \S+ \w+ (\w+) (\d+) (\d+):(\d+):(\d+) (\d+)/
 
-      lock do |f|
+    def update
+      return if FileTest.zero? @filename
+      wf = t = p = nil
+
+      lock( @filename ) {|f|
         begin
           f.each do |line|
-            if /\AFrom /o then
-              if wf then
-                wf.close
+            if /\AFrom / === line then
+              wf.close if wf
+              File.utime t, t, p.filename if t
+
+              p = @real.new_port
+              wf = p.wopen
+              if m = FROMLINE.match(line) then
+                t = Time.local( m[6].to_i, m[1], m[2].to_i,
+                                m[3].to_i, m[4].to_i, m[5].to_i )
+              else
+                t = nil
               end
-              wf = File.open( "#{@tempdir}/#{n}" )
-              n += 1
             else
               wf << line if wf
             end
           end
-          if port then
-            ws.close
-            yield port
-          end
         ensure
           if wf and not wf.closed? then
             wf.close
+            File.utime t, t, p.filename if t
           end
         end
+      }
+      File.truncate @filename, 0
+    end
+  
+    def lock( fname )
+      begin
+        f = File.open( fname )
+        f.flock File::LOCK_EX
+        yield f
+      ensure
+        f.flock File::LOCK_UN
+        f.close
       end
-
-      @real = MhLoader.new( @tempdir )
     end
 
-    def lock
-      begin
-        f = File.open( @filename )
-        begin
-          f.flock File::LOCK_EX
-          yield f
-        ensure
-          f.flock File::LOCK_UN
-        end
-      ensure
-        f.close if f
+    def close_check
+      if @closed then
+        raise ArgumentError, 'accessing already closed mbox'
       end
     end
 
